import logging

from pyramid.interfaces import IRootFactory

from royal import exceptions

log = logging.getLogger(__name__)


def includeme(config):
    config.add_directive('add_resource', add_resource)
    config.add_directive('add_collection', add_collection)
    config.add_directive('add_item', add_item)


category = 'Royal resources'


def add_resource(config, resource_path):
    module_name = '%s.%s' % (config.package_name,
                             resource_path.replace('.', '_'))

    module = config.maybe_dotted(module_name)

    collection_cls = getattr(module, 'Collection', None)
    item_cls = getattr(module, 'Item', None)

    if not any((collection_cls, item_cls)):
        raise exceptions.InvalidConfig('no collection or item class in %s',
                                       module)

    _save_resource_definition(config, resource_path, collection_cls, item_cls)


def add_collection(config, resource_path, cls):
    _save_resource_definition(config, resource_path, cls, None)


def add_item(config, resource_path, cls):
    _save_resource_definition(config, resource_path, None, cls)


def _is_collection(introspectable):
    return introspectable.type_name == 'Collection'


def _is_item(introspectable):
    return introspectable.type_name == 'Item'


def _save_resource_definition(config, resource_path, collection_cls,
                              item_cls):
    try:
        _, name = resource_path.rsplit('.', 1)
    except ValueError:
        # no separator in resource name: it is a root resource
        name = resource_path

    if collection_cls:
        _new_action(config, resource_path, name, collection_cls, 'Collection')

    if item_cls:
        _new_action(config, resource_path, name, item_cls, 'Item')


def _new_action(config, resource_path, name, cls, type_name):
    hierarchy = resource_path.replace('.', '.item.')
    if type_name == 'Item':
        hierarchy += '.item'

    intr = config.introspectable(
        category_name=category,
        discriminator=('royal', hierarchy),
        title=resource_path,
        type_name=type_name,
    )
    intr['name'] = name
    intr['resource_path'] = resource_path
    intr['cls'] = cls
    intr['dotted_name'] = '%s:%s' % (cls.__module__, cls.__name__)

    # Actions must be taken from top-level resource to leaf after
    # root_factory has been set at order=0
    # let's use order parameter:
    # 'app.user.message' collection & item are respectively at order 5 & 6
    # 'app.user' collection & item are respectively at order 3 & 4
    # 'app' collection & item are respectively at order 1 & 2

    order = 1 + (resource_path.count('.') * 2) + (1 if _is_item(intr) else 0)

    config.action(intr.discriminator, register,
                  args=(config, intr), order=order,
                  introspectables=(intr, ))


def register(config, intr):
    parent = find_parent_intr(config.introspector, intr)
    cls = intr['cls']
    name = intr['name']
    cls.children = {}

    if parent is None:
        # root resource: (/users, /bob)
        root_cls = config.registry.queryUtility(IRootFactory)
        intr.relate('root factories', None)
        root_cls.children[name] = cls
        return

    intr.relate(category, parent.discriminator)

    parent_cls = parent['cls']

    if _is_item(intr):
        if _is_item(parent):
            # it's like having /user/1/profile but why not
            parent_cls.children[name] = cls
        else:
            # /users/1, /users/1/photos/123
            parent_cls.item_cls = cls
    else:
        if _is_collection(parent):
            # it's like having /users/photos
            raise exceptions.InvalidConfig(
                'Collection class %s configured with a collection class as'
                ' parent: %s' % (cls, parent_cls)
                )
        # /users/1/photos
        parent_cls.children[name] = cls


def find_parent_intr(introspector, intr):
    hierarchy = intr.discriminator[1]
    try:
        parent_hierarchy, _ = hierarchy.rsplit('.', 1)
    except ValueError:
        return None
    return introspector.get(
        category,
        ('royal', parent_hierarchy),
        )



import os


import logging

from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine
from sqlalchemy.schema import DropConstraint, MetaData


log = logging.getLogger(__name__)


mysql_uri = 'mysql+mysqlconnector://root@localhost/royal_test_example'
engine = None


def setupPackage():
    os.environ['MONGO_URI'] = 'mongodb://localhost'
    os.environ['MONGO_DB_NAME'] = 'royal_example'
    os.environ['MONGO_DB_PREFIX'] = ''

    # sqla extentsion setup.
    global engine

    alembic_config = Config()
    alembic_config.set_main_option('script_location',
                                   'example/ext/sqla/db')
    alembic_config.set_main_option('sqlalchemy.url', mysql_uri)

    engine = create_engine(mysql_uri)

    try:
        command.downgrade(alembic_config, 'base')
    except:
        log.exception("Migration downgrade failed, clearing all tables")
        metadata = MetaData(engine)
        metadata.reflect()
        for table in metadata.tables.values():
            for fk in table.foreign_keys:
                engine.execute(DropConstraint(fk.constraint))
        metadata.drop_all()

    command.upgrade(alembic_config, 'head')



import functools
import venusian


class _declarative_config(object):

    def __init__(self, directive_name, resource_path):
        self.resource_path = resource_path
        self.directive_name = directive_name

    def __call__(self, cls):

        def callback(context, name, cls):
            config = context.config.with_package(info.module)
            directive = getattr(config, self.directive_name)
            directive(self.resource_path, cls, **settings)

        info = venusian.attach(cls, callback)
        settings = {'_info': info.codeinfo}
        return cls


item_config = functools.partial(_declarative_config, 'add_item')
collection_config = functools.partial(_declarative_config, 'add_collection')



from pyramid.events import NewRequest

from royal.declarative import collection_config, item_config
from royal.renderer import renderer_adapter
from royal.resource import Root, Collection, Item
from royal.deserializer import deserializer_config


__all__ = [
    'Collection',
    'deserializer_config',
    'Item',
    'renderer_adapter',
    'Root',
    'collection_config',
    'item_config',
]


def includeme(config):
    config.include('royal.renderer')
    config.include('royal.deserializer')
    config.include('royal.directives')
    config.include('royal.resource')
    config.include('royal.views')

    config.add_subscriber_predicate(
        'request_methods',
        RequestMethodEventPredicate
    )

    config.add_subscriber(
        override_request_method,
        NewRequest,
        request_methods=['POST'],
    )
    config.add_request_method(is_resource_nested, 'is_nested')
    config.commit()


def is_resource_nested(request, resource):
    """Return True if resource is nested in another resource representation.

    It permits to vary a resource representation.
    """
    return request.context not in [resource, resource.parent]


class RequestMethodEventPredicate(object):

    def __init__(self, methods, config):
        self.methods = methods
        self.phash = self.text

    def __call__(self, event):
        return event.request.method in self.methods

    def text(self):
        return 'request_method in %s' % self.methods


def override_request_method(event):
    methods = ['PUT', 'DELETE']
    override = (
        event.request.headers.get('X-HTTP-Method-Override') or
        event.request.POST.get('_method', '').upper()
    )
    if override in methods:
        event.request.method = override



from royal.tests.functional import TestBase

expected_lines = {
    'photos': ('/photos', 'example.resources.photos:Collection',
               'example.resources.photos:Item'),
    'projects': ('/projects', 'example.ext.sqla.resources.projects:Collection',
                 'example.ext.sqla.resources.projects:Item'),
    'users': ('/users', 'example.resources.users:Collection',
              'example.resources.users:Item'),
    'users.followers': ('/users/(id)/followers',
                        'example.resources.users:UserFollowers',
                        'example.resources.users:Follower'),
    'users.photos': ('/users/(id)/photos',
                     'example.resources.users_photos:Collection'),
    }


class Test(TestBase):

    def test_presources(self):
        from royal.presources import get_pretty_resource_list
        self.config.commit()
        registry = self.config.registry
        result = get_pretty_resource_list(registry).splitlines()
        result.pop(1)  # header underline

        expected = expected_lines.copy()
        for line in result:
            terms = line.split()

            if terms[0] not in expected_lines:
                continue

            columns = expected_lines[terms[0]]

            for term in columns:
                self.assertIn(term, term)

            expected.pop(terms[0])

        self.assertEqual(expected, {})



from sqlalchemy.exc import IntegrityError

from royal.tests.functional import TestBase

PR_ID = 3
PROJECT_NAME = 'TEST_P3'


class Test(TestBase):

    def setUpModels(self):
        from example.ext.sqla.model import Project
        self.session.add(Project(id=PR_ID, name=PROJECT_NAME))

    def tearDownModels(self):
        from example.ext.sqla.model import Project
        self.session.query(Project).delete()

    def test_project_index(self):
        response = self.app.get('/projects').json
        self.assertIn('items', response)
        self.assertIn('links', response)
        self.assertEqual(len(response['items']), 1)

        project = response['items'][0]
        self.assertEqual(project['id'], PR_ID)
        self.assertEqual(project['name'], PROJECT_NAME)

    def test_project_create(self):
        result = self.app.post('/projects', {'name': 'test_proj'}).json
        new_proj_id = result['id']
        self.app.get('/projects/%s' % new_proj_id)

    def test_project_show(self):
        self.app.get('/projects/999', status=404)

        expected = {'id': PR_ID,
                    'links': {'self': 'http://localhost/projects/%s/' % PR_ID},
                    'name': u'TEST_P3'}
        result = self.app.get('/projects/%s' % PR_ID).json
        self.assertEqual(result, expected)

    def test_project_delete(self):
        self.app.delete('/projects/%s' % PR_ID)
        self.app.delete('/projects/%s' % PR_ID, status=404)

    def test_project_update(self):
        self.app.patch('/projects/999', {'name': 'new_name'}, status=404)
        self.app.patch('/projects/%s' % PR_ID, {'name': 'new_name'})

        expected = {'id': PR_ID,
                    'links': {'self': 'http://localhost/projects/%s/' % PR_ID},
                    'name': u'new_name'}
        result = self.app.get('/projects/%s' % PR_ID).json
        self.assertEqual(result, expected)

    def test_project_create_duplicate(self):

        with self.assertRaises(IntegrityError):
            self.app.post('/projects', {'name': PROJECT_NAME})



# -*- coding: utf-8
"""List all royal resources.
"""

import argparse
import logging
import functools
from collections import OrderedDict

from pyramid.paster import bootstrap

from royal.directives import category

log = logging.getLogger(__name__)


def main():  # pragma no cover
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('config_uri',
                        help='PasteDeploy config file to use.')
    args = parser.parse_args()

    env = bootstrap(args.config_uri)
    print get_pretty_resource_list(env['registry'])


def get_pretty_resource_list(registry):
    introspector = registry.introspector
    intr_seq = introspector.get_category(category)

    introspectables = sorted([d['introspectable'] for d in intr_seq],
                             cmp=lambda x, y: cmp(x.title, y.title))

    resources = OrderedDict()
    for intr in introspectables:
        if intr.title in resources:
            resource = resources[intr.title]
        else:
            resource = resources[intr.title] = {
                'name': intr['name'],
                'pattern': '/' + intr.title.replace('.', '/(id)/'),
                'collection': None,
                'item': None,
            }

        resource[intr.type_name.lower()] = intr['dotted_name']

    return format(resources)


def format(resources):
    max_path_len = len(max(resources.keys(), key=len))
    max_pattern_len = len(max([r['pattern'] for k, r in resources.items()],
                              key=len))
    max_collection_len = len(max([r['collection']
                                  for k, r in resources.items()
                                  if r['collection']],
                                 key=len))
    max_item_len = len(max([r['item'] for k, r in resources.items()
                            if r['item']],
                           key=len))

    get_line = functools.partial(get_formatted_line,
                                 max_path_len + 1,
                                 max_pattern_len + 1,
                                 max_collection_len + 1,
                                 max_item_len + 1)

    header = get_line('Resource', 'Pattern', 'Collection', 'Item')

    underline = get_line('=' * (max_path_len + 1),
                         '=' * (max_pattern_len + 1),
                         '=' * (max_collection_len + 1),
                         '=' * (max_item_len + 1))
    body = ''
    for path, info in resources.items():
        body += get_line(path, info['pattern'], info['collection'],
                         info['item'])

    return header + underline + body


def get_formatted_line(column1_len, column2_len, column3_len, column4_len,
                       column1, column2, column3, column4):
    fmt = '{column1:{fill}{align}{column1_len}}'
    fmt += '{column2:{fill}{align}{column2_len}}'
    fmt += '{column3:{fill}{align}{column3_len}}'
    fmt += '{column4:{align}}\n'
    if column3 is None:
        column3 = ''
    if column4 is None:
        column4 = ''
    params = dict(column1=column1, column2=column2, column3=column3,
                  column4=column4, fill=' ', align='<',
                  column1_len=column1_len + 1,
                  column2_len=column2_len + 1,
                  column3_len=column3_len + 1,
                  column4_len=column4_len + 1)
    return fmt.format(**params)



import setuptools

setuptools.setup(
    setup_requires=['d2to1'],
    d2to1=True,
    test_suite='nose.collector',
    entry_points={
        'console_scripts': [
            'presources = royal.presources:main',
        ],
    },
)
